##\internal
"""@package graphlab.toolkits

This module defines the (internal) utility functions used by the toolkits.
"""

from array import array as _array
import json
from graphlab.data_structures.sframe import SArray as _SArray
from graphlab.data_structures.sframe import SFrame as _SFrame
from graphlab.data_structures.sgraph import SGraph as _SGraph
from graphlab.data_structures.sgraph import Vertex as _Vertex
from graphlab.data_structures.sgraph import Edge as _Edge
from graphlab.cython.cy_sarray import UnitySArrayProxy
from graphlab.cython.cy_sframe import UnitySFrameProxy
from graphlab.cython.cy_graph import UnityGraphProxy
from graphlab.toolkits._main import ToolkitError
import logging as _logging


_proxy_map = {UnitySFrameProxy: (lambda x: _SFrame(_proxy=x)),
              UnitySArrayProxy: (lambda x: _SArray(_proxy=x)),
              UnityGraphProxy: (lambda x: _SGraph(_proxy=x))}

def _add_docstring(format_dict):
  """
  Format a doc-string on the fly.
  @arg format_dict: A dictionary to format the doc-strings
  Example:

    @add_docstring({'context': __doc_string_context})
    def predict(x):
      '''
      {context}
        >> model.predict(data)
      '''
      return x
  """
  def add_docstring_context(func):
      def wrapper(*args, **kwargs):
          return func(*args, **kwargs)
      wrapper.__doc__ = func.__doc__.format(**format_dict)
      return wrapper
  return add_docstring_context


def _SGraphFromJsonTree(json_str):
    """
    Convert the Json Tree to SGraph
    """
    g = json.loads(json_str)
    vertices = [_Vertex(x['id'],
                dict([(str(k), v) for k, v in x.iteritems() if k != 'id']))
                                                      for x in g['vertices']]
    edges = [_Edge(x['src'], x['dst'],
             dict([(str(k), v) for k, v in x.iteritems() if k != 'src' and k != 'dst']))
                                                      for x in g['edges']]
    sg = _SGraph().add_vertices(vertices)
    if len(edges) > 0:
        sg = sg.add_edges(edges)
    return sg

def _toolkit_repr_print(model, fields, width=20):
    """
    Display a toolkit repr according to some simple rules.

    @arg model : GraphLab model
    @arg fields: List of tuples (display_name, field_name)

    Example:

        model_fields = [
            ("L1 penalty", 'l1_penalty'),
            ("L2 penalty", 'l2_penalty'),
            ("Examples", 'num_examples'),
            ("Features", 'num_features'),
            ("Coefficients", 'num_coefficients')]

        solver_fields = [
            ("Solver", 'solver'),
            ("Solver iterations", 'training_iterations'),
            ("Solver status", 'training_solver_status'),
            ("Training time (sec)", 'training_time')]

        training_fields = [
            ("Log-likelihood", 'training_loss')]

        fields = [model_fields, solver_fields, training_fields]:
    """
    key_str = "{:<{}}: {}"

    ret = []
    ret.append(key_str.format("Class", width, model.__class__.__name__))
    for tranche_fields in fields:
        for k, v in tranche_fields:
            value = model.get(v)
            if isinstance(value, float):
                try:
                    value = round(value, 4)
                except:
                    pass
            ret.append(key_str.format(k, width, value))
        ret.append("")
    return '\n'.join(ret)

def _map_unity_proxy_to_object(value):
    """
    Map returning value, if it is unity SFrame, SArray, map it
    """
    vtype = type(value)
    if vtype in _proxy_map:
        return _proxy_map[vtype](value)
    elif vtype == list:
        return [_map_unity_proxy_to_object(v) for v in value]
    elif vtype == dict:
        return {k:_map_unity_proxy_to_object(v) for k,v in value.items()}
    else:
        return value

def _toolkits_select_columns(dataset, columns):
    """
    Same as select columns but redirect runtime error to ToolkitError.
    """
    try:
        return dataset.select_columns(columns)
    except RuntimeError as e:
        raise ToolkitError, str(e)

def _raise_error_if_column_exists(dataset, column_name = 'dataset',
                            dataset_variable_name = 'dataset',
                            column_name_error_message_name = 'column_name'):
    """
    Check if a column exists in an SFrame with error message.
    """
    err_msg = 'The SFrame {0} must contain the column {1}.'.format(
                                                dataset_variable_name,
                                             column_name_error_message_name)
    if column_name not in dataset:
      raise ToolkitError, str(err_msg)

def _check_categorical_option_type(option_name, option_value, possible_values):
    """
    Check whether or not the requested option is one of the allowed values.
    """
    err_msg = 'The provided value for {0} is not valid. '.format(option_name)
    err_msg += '{0} should be changed to one of the following:'.format(option_value)
    err_msg += ', '.join(possible_values)
    if option_value not in possible_values:
        raise ToolkitError, err_msg

def _raise_error_if_not_sarray(dataset, variable_name="SFrame"):
    """
    Check if the input is an SArray. Provide a proper error
    message otherwise.
    """
    err_msg = "Input %s is not an SArray."
    if not isinstance(dataset, _SArray):
      raise ToolkitError, err_msg % variable_name

def _raise_error_if_not_sframe(dataset, variable_name="SFrame"):
    """
    Check if the input is an SFrame. Provide a proper error
    message otherwise.
    """
    err_msg = "Input %s is not an SFrame. If it is a Pandas dataframe,"
    err_msg += " you may use the to_sframe() function to convert it to an SFrame."

    if not isinstance(dataset, _SFrame):
      raise ToolkitError, err_msg % variable_name

def _raise_error_if_sframe_empty(dataset, variable_name="SFrame"):
    """
    Check if the input is empty.
    """
    err_msg = "Input %s either has no rows or no columns. A non-empty SFrame "
    err_msg += "is required."

    if dataset.num_rows() == 0 or dataset.num_cols() == 0:
        raise ToolkitError, err_msg % variable_name

def _raise_error_if_not_iterable(dataset, variable_name="SFrame"):
    """
    Check if the input is iterable.
    """
    err_msg = "Input %s is not iterable: hasattr(%s, '__iter__') must be true."

    if not hasattr(dataset, '__iter__'):
        raise ToolkitError, err_msg % variable_name


def _raise_error_evaluation_metric_is_valid(metric, allowed_metrics):
    """
    Check if the input is an SFrame. Provide a proper error
    message otherwise.
    """

    err_msg = "Evaluation metric '%s' not recognized. The supported evaluation"
    err_msg += " metrics are (%s)."

    if metric not in allowed_metrics:
      raise ToolkitError, err_msg % (metric,
                          ', '.join(map(lambda x: "'%s'" % x, allowed_metrics)))

def _validate_array_type(dataset, col_name):
    """
    Check that each element of the array contains an equal number of numeric types.
    The check is performed on the first 100 or all rows of the column, whichever
    is smaller.

    Parameters
    ----------
    dataset: SFrame
        Input dataset.

    col_name: str
        Name of the column to check.

    Returns
    -------
    out: bool
        If True, 'col_name' is a valid array-type feature.
    """
    is_valid = True
    nsamples = min(dataset.num_rows(), 100)
    m = 0
    for i in range(nsamples):
        if i == 0:
            m = len(dataset[col_name][i])

        if len(dataset[col_name][i]) != m:
            is_valid = False
            _logging.warning("Column " + col_name +
                " is excluded as a feature because row " + str(i) +
                " contains an array of a different length. All arrays in the " +
                " column must have equal length.\n")
            break

        # Now verify that each element in the array is a number
        for j, elem in enumerate(dataset[col_name][i]):
            if type(elem) not in [float, int]:
                is_valid = False
                _logging.warning("Column " + col_name +
                    " is excluded as a feature because row " + str(i) + 
                    " element " + str(j) + "is an element of non-numeric type." +
                    " Only numeric array elements (int or float) are accepted" +
                    " as valid features.\n")
                break

        if not is_valid:
            break

    return is_valid

def _validate_dict_type(dataset, col_name):
    """
    Check that each dictionary in the specified column contains only numeric values.
    The check is performed on the first 100 or all rows of the column, whichever
    is smaller.

    Parameters
    ----------
    dataset: SFrame
        Input dataset.

    col_name: str
        Name of the column to check.

    Returns
    -------
    out: bool
        If True, 'col_name' is a valid dict-type feature.
    """
    is_valid = True
    nsamples = min(dataset.num_rows(), 100)
    for i in range(nsamples):
        for k, v in dataset[col_name][i].iteritems():
            if type(v) not in [float, int]:
                is_valid = False
                _logging.warning("Column " + col_name +
                    " is excluded as a feature because row " + str(i) + " key " + str(k) +
                    " has a value of non-numeric type. Only numeric value types (int or float) are accepted as valid features.\n")
                break

        if not is_valid:
            break

    return is_valid

def _select_valid_features(dataset, features, valid_feature_types,
                           target_column=None):
    """
    Utility function for selecting columns of only valid feature types.

    Parameters
    ----------
    dataset: SFrame
        The input SFrame containing columns of potential features.

    features: list[str]
        List of feature column names.  If None, the candidate feature set is
        taken to be all the columns in the dataset.

    valid_feature_types: list[type]
        List of Python types that represent valid features.  If type is array.array,
        then an extra check is done to ensure that the individual elements of the array
        are of numeric type.  If type is dict, then an extra check is done to ensure
        that dictionary values are numeric.

    target_column: str
        Name of the target column.  If not None, the target column is excluded
        from the list of valid feature columns.

    Returns
    -------
    out: list[str]
        List of valid feature column names.  Warnings are given for each candidate
        feature column that is excluded.

    Examples
    --------
    # Select all the columns of type `str` in sf, excluding the target column named
    # 'rating'
    >>> valid_columns = _select_valid_features(sf, None, [str], target_column='rating')

    # Select the subset of columns 'X1', 'X2', 'X3' that has dictionary type or defines
    # numeric array type
    >>> valid_columns = _select_valid_features(sf, ['X1', 'X2', 'X3'], [dict, array.array])
    """
    if features is not None:
        if not hasattr(features, '__iter__'):
            raise TypeError("Input 'features' must be an iterable type.")

        if not all([isinstance(x, str) for x in features]):
            raise TypeError("Input 'features' must contain only strings.")

    ## Extract the features and labels
    if features is None:
        features = dataset.column_names()

    col_type_map = {
        col_name: col_type for (col_name, col_type) in
        zip(dataset.column_names(), dataset.column_types()) }

    valid_features = []
    for col_name in features:

        if col_name == target_column:
            _logging.warning("Excluding target column " + target_column +\
                             " as a feature.")

        else:
            include_flag = True     

            if col_type_map[col_name] not in valid_feature_types:
                include_flag = False

            else:
                if col_type_map.get(col_name) == _array:
                    include_flag = _validate_array_type(dataset, col_name)
                if col_type_map.get(col_name) == dict:
                    include_flag = _validate_dict_type(dataset, col_name)

            if include_flag is True:
                valid_features.append(col_name)
            else:
                _logging.warning("Column " + col_name +
                    " is excluded as a feature due to invalid column type: " +
                    str(col_type_map.get(col_name)) + ".\n")


    if len(valid_features) == 0:
        raise ValueError("The dataset does not contain any valid feature columns. " +
            "Accepted feature types are " + str(valid_feature_types) + ".")

    return valid_features


def _numeric_param_check_range(variable_name, variable_value, range_bottom, range_top):
    """
    Checks if numeric parameter is within given range
    """
    err_msg = "%s must be between %i and %i"

    if variable_value < range_bottom or variable_value > range_top:
        raise ToolkitError, err_msg % (variable_name, range_bottom, range_top)
